


#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "itkImage.h"
#include "itkImportImageFilter.h"
#include "itkWatershedImageFilter.h"

#include "itkImageFileWriter.h"
#include "itkImageFileReader.h"
#include "itkVectorCastImageFilter.h"
#include "itkUnaryFunctorImageFilter.h"
#include "itkRGBPixel.h"
#include "itkScalarToRGBPixelFunctor.h"
#include "itkRegionOfInterestImageFilter.h"
#include "itkExtractImageFilter.h"
#include "itkMedianImageFilter.h"
#include "itkCurvatureFlowImageFilter.h"
#include "tomo.h"
extern void input_ctrl_struct_values(char * filename, ctrl_struct *Struct_Ptr);
extern void init_ctrl (Ctrl * cp);
extern void free_ctrl (Ctrl * cp);
extern TOMODATA * readdata(const char * fname,int datasize[3]);

void getMinf (FILE * fp){
 #ifdef RCAPRINTMEM
 static struct mallinfo Minf;
 Minf = mallinfo();
 fprintf(fp,"mallinfo arena %i hblkhd %i ordblks %i uordblks %i fordblks %i\n",Minf.arena,Minf.hblkhd,Minf.ordblks,Minf.uordblks,Minf.fordblks);
 #endif
 return;
}

int main(int argc, char ** argv) {
   TOMODATA * imagedata;
   Ctrl * ctrl;
   FILE * logfile;
   
  if( argc < 2 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << "  ctrl.in file" << std::endl; 
    return 1;
    }

/* read the file using Robert Atwood's routines */
   ctrl = (Ctrl *) calloc(1,sizeof(Ctrl));
   init_ctrl (ctrl);
   input_ctrl_struct_values(*(argv+1),ctrl);
   imagedata = readdata(ctrl->infile,ctrl->datasize);

  
  typedef tomo_t PixelType;

  typedef itk::Image< PixelType, 3 > ScalarImageType;
  typedef itk::RGBPixel<unsigned char>   RGBPixelType;
  typedef itk::Image<RGBPixelType, 3>    RGBImageType;
  typedef itk::Image<RGBPixelType, 2 >   RGBImageSlice;
  typedef itk::Image<unsigned long, 3>   LabeledImageType;
  typedef itk::Image<unsigned long, 2>   LabeledImageSlice;
  typedef itk::WatershedImageFilter<ScalarImageType> WatershedFilterType;

  typedef itk::ImportImageFilter< PixelType, 3 >   ImportFilterType;

  // definitions for the colorizer for watershed output 
  typedef itk::Functor::ScalarToRGBPixelFunctor<unsigned long> ColorMapFunctorType;
  typedef itk::UnaryFunctorImageFilter<LabeledImageType, RGBImageType, ColorMapFunctorType> ColorMapFilterType;
  typedef itk::UnaryFunctorImageFilter<LabeledImageSlice, RGBImageSlice, ColorMapFunctorType> ColorMapFilterSlice;

  // definitions for writing the image
  typedef itk::ImageFileWriter< RGBImageType > WriterType;
  typedef itk::ImageFileWriter< LabeledImageType > LabeledWriterType;

  /* definitions needed for slice extrat and write */
  typedef itk::ImageFileWriter< RGBImageSlice > SliceWriterType;
  typedef itk::ExtractImageFilter<LabeledImageType,LabeledImageSlice> SliceExtractType;

  /* definition for region extractor */
  typedef itk::ExtractImageFilter<ScalarImageType,ScalarImageType> RegionExtractType;

/* definitions for pre-filter */
  typedef itk::CurvatureFlowImageFilter<ScalarImageType,ScalarImageType> PreFilterType;

/* definitions for post-filter */
  typedef itk::MedianImageFilter<LabeledImageSlice,LabeledImageSlice> SliceMedian;
  typedef itk::MedianImageFilter<LabeledImageType,LabeledImageType> VolMedian;


/* objects for importing the data */
  ImportFilterType::Pointer importFilter = ImportFilterType::New();      
  ImportFilterType::SizeType  size;
  ImportFilterType::IndexType start;
  ImportFilterType::RegionType region;

  /* objects needed for filtering */
  WatershedFilterType::Pointer watershed = WatershedFilterType::New();

  ColorMapFilterSlice::Pointer colormapperslice = ColorMapFilterSlice::New();



  SliceMedian::Pointer slicemedian = SliceMedian::New();
  LabeledImageSlice::SizeType slm_rad;


/* objects needed for slice extract */
  SliceExtractType::Pointer slicer = SliceExtractType::New();
  /* settign up the sliced region */
     LabeledImageType::RegionType inputRegion ;
     LabeledImageType::SizeType inputRegionSize ;
     LabeledImageType::IndexType inputRegionIndex ;

  /* objects neede for writing out the data */
  SliceWriterType::Pointer slicewriter = SliceWriterType::New();

  ScalarImageType::Pointer filteredImage;


  unsigned int sliceNumber;
  char slicename[255];
  char filename[255];
  int i,j,istart,iend,jstart,jend;
  float level,thresh;
  float pre_step;
  unsigned int pre_iter;
  float wstart,wend,wstep,tstart,tend,tstep;


  double origin[ 3 ];
  double spacing[ 3 ];
  const bool importImageFilterWillOwnTheBuffer = false;

  /* the import filter does NOT own th buffer as it was created by malloc */

  /* create a log file to put messages in */
  snprintf(filename,255,"%s.log",ctrl->outfile);
  logfile = fopen(filename,"w");

  /* import the data from a raw file with control file header */
  /* imported to imagedata structure */

  size[0]  = imagedata->tsizes[0];  // size along X
  size[1]  = imagedata->tsizes[1];  // size along Y
  size[2]  = imagedata->tsizes[2];  // size along Z

  start.Fill( 0 );

  region.SetIndex( start );
  region.SetSize(  size  );
  importFilter->SetRegion( region );

  origin[0] = imagedata->origin[0];    // X coordinate 
  origin[1] = imagedata->origin[1];    // Y coordinate
  origin[2] = imagedata->origin[2];    // Z coordinate

  importFilter->SetOrigin( origin );

  /* voxel units */

  spacing[0] = 1.0;    // along X direction 
  spacing[1] = 1.0;    // along Y direction
  spacing[2] = 1.0;    // along Z direction

  importFilter->SetSpacing( spacing );
  importFilter->SetImportPointer( imagedata->tdata, imagedata->ndata, 
                                  importImageFilterWillOwnTheBuffer );

/* the import filter does NOT own th buffer as it was created by malloc */
/* not c++ new */

	  getMinf(stdout);
	  getMinf(logfile);

  /* initial default */
  watershed->SetLevel( 0.9 );
  watershed->SetThreshold( 0.05 );


  /* possible post filter using median */
  slm_rad[0] = 3;
  slm_rad[1] = 3;
  slicemedian->SetRadius(slm_rad);



  /* which slice to print for inspection */
  sliceNumber = ctrl->slicenum;

  /* set up the looping, DECREMENT wlevel as it is more efficient */
  /* with the itk algorithm */
  wstart = ctrl->wtop;
  wend = ctrl->wbot;
  wstep = (wstart -wend)/(ctrl->nloop);
  tstart = ctrl->tbot;
  tend = ctrl->ttop;
  tstep = (tend - tstart)/(ctrl->nloop);


  printf("wstart %f wend %f wstep %f\n",wstart,wend,wstep);
  printf("tstart %f tend %f tstep %f\n",tstart,tend,tstep);
  printf("nloop %i\n",ctrl->nloop);

  fprintf(logfile,"wstart %f wend %f wstep %f\n",wstart,wend,wstep);
  fprintf(logfile,"tstart %f tend %f tstep %f\n",tstart,tend,tstep);
  fprintf(logfile,"nloop %i\n",ctrl->nloop);
  
/* use just one fixed paramete if loop is not selected */
i=ctrl->wshd_level;
j=ctrl->wshd_thr;

if (ctrl->wloop){
    istart = 0;
    iend = ctrl->nloop;
}else{
    istart = i;
    iend = i;       
}
if (ctrl->tloop){
    jstart = 0;
    jend = ctrl->nloop;
}else{
    jstart = j;
    jend = j;       
}
   
 
  


           getMinf(stdout);
           getMinf(logfile);

  { /* block of existance for prefilter */
        PreFilterType::Pointer prefilter = PreFilterType::New();
        RegionExtractType::Pointer regionExtractor = RegionExtractType::New();


           getMinf(stdout);
           getMinf(logfile);
        /* choose selected region or else use the whole image */
        if (ctrl->use_chop == 1){
           printf("Using selected region \n");
           fprintf(logfile,"Using selected region \n");

           /* set up a region */
           inputRegion = importFilter->GetOutput()->GetLargestPossibleRegion();
           inputRegionSize = inputRegion.GetSize();
           inputRegionIndex = inputRegion.GetIndex(); 

           for(i=0;i<3;i++){
               inputRegionIndex[i] = ctrl->choplimits[i];
               inputRegionSize[i] = ctrl->choplimits[i+3]-ctrl->choplimits[i];
           }

           inputRegion.SetSize(inputRegionSize);
           inputRegion.SetIndex(inputRegionIndex);
           regionExtractor->SetExtractionRegion(inputRegion);
           regionExtractor->SetInput(importFilter->GetOutput());

           prefilter->SetInput( regionExtractor->GetOutput());
        }else{
           printf("Using whole image  \n");
           fprintf(logfile,"Using whole image  \n");
          prefilter->SetInput(importFilter->GetOutput());
        }

           getMinf(stdout);
           getMinf(logfile);

           /* set up any prefiltering of the input map */
           pre_step=ctrl->pre_step;
           pre_iter=ctrl->pre_iter;
           prefilter->SetTimeStep(pre_step);
           prefilter->SetNumberOfIterations(pre_iter);

        printf("Updating prefilter ... using pre_step %f pre_iter %i\n",pre_step,pre_iter);
        fprintf(logfile,"Updating prefilter ... using pre_step %f pre_iter %i\n",pre_step,pre_iter);
        try
        {
           prefilter->Update();
        }
        catch( itk::ExceptionObject & exp ) 
        {
          std::cerr << "Exception caught 001 prefilter !" << std::endl;
          std::cerr << "Exiting !" << std::endl;
          std::cerr << exp << std::endl;
          exit(0);
          /* exit, otherwise images get written anyways using the wrong output */
          /* causing erroneous conclusions! */
        }
        filteredImage = prefilter->GetOutput();
        try{
        filteredImage->Update();
        }catch( itk::ExceptionObject & exp ) {
          std::cerr << "Exception caught 001 filteredImage !" << std::endl;
          std::cerr << "Exiting !" << std::endl;
          std::cerr << exp << std::endl;
          exit(0);
          /* exit, otherwise images get written anyways using the wrong output */
          /* causing erroneous conclusions! */
        }
        printf("finished Updating prefilter ... \n");
        fprintf(logfile,"finished Updating prefilter ...\n");

  } /*end  block of existance for prefilter */

  /* can get rid of the original image now */
	  getMinf(stdout);
	  getMinf(logfile);

   printf("freeing input image...\n");
   fprintf(logfile,"freeing input image ...\n");
   free(imagedata->tdata);
   free(imagedata->filename);
   free(imagedata);

           getMinf(stdout);
           getMinf(logfile);


  /* processing pipeline for watershed parameter loop */
  watershed->SetInput( filteredImage );


	  getMinf(stdout);
	  getMinf(logfile);

#ifdef INITIAL
  printf("Updating initial watershed ...\n");
  fprintf(logfile,"Updating initial watershed ...\n");
  try{
  watershed->Update();
  }catch(itk::ExceptionObject & exp ) {
    std::cerr << "Exception caught 001 !" << std::endl;
    std::cerr << "Exiting !" << std::endl;
    std::cerr << exp << std::endl;
    exit(0);
    /* exit, otherwise images get written anyways using the wrong output */
    /* causing erroneous conclusions! */
  }
#endif

  // watershed->UpdateLargestPossibleRegion();

	  getMinf(stdout);
	  getMinf(logfile);
   
#ifdef INITIAL
/* set up a slice */
  inputRegion = watershed->GetOutput()->GetLargestPossibleRegion();
  inputRegionSize = inputRegion.GetSize();
  inputRegionIndex = inputRegion.GetIndex(); 

  if (sliceNumber >= (inputRegionSize[2] + inputRegionIndex[2]) || sliceNumber < inputRegionIndex[2] ){
     sliceNumber = (int)(inputRegionIndex[2]);
     fprintf(stderr,"WARNING: slicenumber out of range, resetting to %i\n" , sliceNumber);
     fprintf(logfile,"WARNING: slicenumber out of range, resetting to %i\n" , sliceNumber);
  }

  inputRegionSize[2] = 0;
  inputRegionIndex[2] = sliceNumber; 
  inputRegion.SetSize(inputRegionSize);
  inputRegion.SetIndex(inputRegionIndex);
  slicer->SetExtractionRegion(inputRegion);
#endif

/* add slicer to the pipe line */
  slicer->SetInput ( watershed->GetOutput());
  colormapperslice->SetInput(slicer->GetOutput());
  slicewriter->SetInput(colormapperslice->GetOutput());
  


/* loop through the threshold values */
for (j=jstart;j<=jend;j++){
     thresh = tstart + (tstep * (float)(j)) ;
     printf("setting watershed threshold %f\n",thresh);
     fprintf(logfile,"setting watershed threshold %f\n",thresh);
     watershed->SetThreshold(thresh);


   for(i=istart;i<=iend;i++){
   /* w-level should decrement if looping, instead of increment */
   /* according to the itk guide, this avoids redoing the watersheds*/

        level = wstart - (wstep * (float)(i)) ;
        watershed->SetLevel( level );

        printf("setting watershed level %f\n",level);
        fprintf(logfile,"setting watershed level %f\n",level);


        printf("Iteration %i %i : l=%f t=%f\n",i,j,level,thresh);
        fprintf(logfile,"Iteration %i %i : l=%f t=%f\n",i,j,level,thresh);

      /* set up the slice writer */
        snprintf(slicename,255,"slice%03i_l%03i_t%03i_%s.png",sliceNumber,i,j,ctrl->outfile);
        /* temporary debug- write over the same file to avod creating 100000's of files while testing loop*/
        //snprintf(slicename,255,"slice.png");
        slicewriter->SetFileName(slicename);
        
        getMinf(stdout);
        getMinf(logfile);

      /* process the slice */
        printf("Updating filter pipeline and printing slice:\n%s\n",slicename);
        fprintf(logfile,"Updating filter pipeline and printing slice\n%s\n",slicename);
        //for debugging, separate the updates into try/catch blocks 
        try
          {
          watershed->Update();
          }
        catch( itk::ExceptionObject & exp ) 
          {
             std::cerr << "Exception caught 01 wshd !" << std::endl;
             std::cerr << "Exiting !" << std::endl;
             std::cerr << exp << std::endl;
             exit(0);
             /* exit, otherwise images get written anyways using the wrong output */
             /* causing erroneous conclusions! */
          }
#ifndef INITIAL
/* set up a slice */
  inputRegion = watershed->GetOutput()->GetLargestPossibleRegion();
  inputRegionSize = inputRegion.GetSize();
  inputRegionIndex = inputRegion.GetIndex(); 

  if (sliceNumber >= (inputRegionSize[2] + inputRegionIndex[2]) || sliceNumber < inputRegionIndex[2] ){
     sliceNumber = (int)(inputRegionIndex[2]);
     fprintf(stderr,"WARNING: slicenumber out of range, resetting to %i\n" , sliceNumber);
     fprintf(logfile,"WARNING: slicenumber out of range, resetting to %i\n" , sliceNumber);
  }

  inputRegionSize[2] = 0;
  inputRegionIndex[2] = sliceNumber; 
  inputRegion.SetSize(inputRegionSize);
  inputRegion.SetIndex(inputRegionIndex);
  slicer->SetExtractionRegion(inputRegion);
#endif

        //expanded into several try.catch blocks to see exactly where
        // the exeption occurs when looping large volume

        try
          {
          slicer->Update();
          }
        catch( itk::ExceptionObject & exp ) 
          {
             std::cerr << "Exception caught 01 slicer !" << std::endl;
             std::cerr << "Exiting !" << std::endl;
             std::cerr << exp << std::endl;
             exit(0);
             /* exit, otherwise images get written anyways using the wrong output */
             /* causing erroneous conclusions! */
          }
        try
          {
          colormapperslice->Update();
          }
        catch( itk::ExceptionObject & exp ) 
          {
             std::cerr << "Exception caught 01 colormapperslice !" << std::endl;
             std::cerr << "Exiting !" << std::endl;
             std::cerr << exp << std::endl;
             exit(0);
             /* exit, otherwise images get written anyways using the wrong output */
             /* causing erroneous conclusions! */
          }
        try
          {
          slicewriter->Update();
          }
        catch( itk::ExceptionObject & exp ) 
          {
             std::cerr << "Exception caught 01 slicewriter !" << std::endl;
             std::cerr << "Exiting !" << std::endl;
             std::cerr << exp << std::endl;
             exit(0);
             /* exit, otherwise images get written anyways using the wrong output */
             /* causing erroneous conclusions! */
          }
        getMinf(stdout);
        getMinf(logfile);

    } /* i */
    fflush(logfile);
} /* j */

/* process the volume */
if (ctrl->process){

  /* objects only needed for processing the volume after watershedding */
  /* created only in the section for processing the volume */

  /*
  ColorMapFilterType::Pointer colormapper = ColorMapFilterType::New();
  VolMedian::Pointer volmedian = VolMedian::New();
  WriterType::Pointer writer = WriterType::New();
  */

  LabeledImageType::SizeType vol_rad;
  LabeledWriterType::Pointer lwriter = LabeledWriterType::New();

  /*
  vol_rad[0] = 1;
  vol_rad[1] = 1;
  vol_rad[2] = 1;
  volmedian->SetRadius(vol_rad);

  colormapper->SetInput(watershed->GetOutput() );

  writer->SetInput( colormapper->GetOutput() );
  writer->SetFileName( ctrl->outfile );
  */

  lwriter->SetInput(watershed->GetOutput());

  snprintf(slicename,255,"lab_l%03i_t%03i_%s",i-1,j-1,ctrl->outfile);
  printf("..writing watershed volume file %s ...\n",slicename);
  lwriter->SetFileName( slicename );
  try
    {
    lwriter->Update();
    // writer->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught 03 !" << std::endl;
    std::cerr << exp << std::endl;
    }
/*
   printf("...3d Median ...\n");
   volmedian->SetInput(watershed->GetOutput());
   colormapper->SetInput(volmedian->GetOutput());
   sprintf(slicename,"med_%s",ctrl->outfile);
   writer->SetFileName( slicename );
  try
    {
    writer->Update();
    }
  catch( itk::ExceptionObject & exp ) 
    {
    std::cerr << "Exception caught 04 !" << std::endl;
    std::cerr << exp << std::endl;
    }
*/
}
    


  fclose(logfile);
  return 0;
}
  

